#ifndef EZC3D_MATH_H
#define EZC3D_MATH_H
///
/// \file ezc3dMath.h
/// \brief Include all math
/// \author Pariterre
/// \version 1.0
/// \date March 25th, 2020
///

#include "math/Matrix.h"
#include "math/Matrix33.h"
#include "math/Matrix66.h"
#include "math/Vector3d.h"
#include "math/Vector6d.h"

#endif
