#ifndef EZC3D_ALL_H
#define EZC3D_ALL_H
///
/// \file ezc3d_all.h
/// \brief Include all the necessary files to enjoy ezc3ding
/// \author Pariterre
/// \version 1.0
/// \date March 27th, 2020
///


#include "math/ezc3dMath.h"
#include "Header.h"
#include "Parameters.h"
#include "Data.h"

#include "modules/ForcePlatforms.h"

#endif
